use crate::*;

use std::ops::Deref;
use subxt::tx::{BaseExtrinsicParamsBuilder, PairSigner};

type SessionKeys = [u8; 128];

/// define smith subcommands
#[derive(Clone, Default, Debug, clap::Parser)]
pub enum Subcommand {
	/// Emit a smith certification
	Cert { to: u32 },
	/// go online
	GoOnline,
	#[default]
	/// go offline
	GoOffline,
	/// Rotate and set session keys
	UpdateKeys,
	/// set sudo keys
	SudoSetKey {
		new_key: AccountId,
	},
	/// List upcoming expirations that require an action
	ShowExpire {
		/// Show certs that expire within less than this number of blocks
		#[clap(short, long, default_value_t = 100800)]
		blocks: u32,
		/// Show authorities that should rotate keys within less than this number of sessions
		#[clap(short, long, default_value_t = 100)]
		sessions: u32,
	},
	/// List online authorities
	ShowOnline,
}

/// handle smith commands
pub async fn handle_command(data: Data, command: Subcommand) -> anyhow::Result<()> {
	let mut data = data.build_client().await?.build_indexer().await?;
	match command {
		Subcommand::GoOnline => {
			commands::smith::go_online(&data).await?;
		}
		Subcommand::GoOffline => {
			commands::smith::go_offline(&data).await?;
		}
		Subcommand::Cert { to } => {
			data = data.build_keypair().fetch_idty_index().await?;
			commands::smith::cert(&data, to).await?
		}
		Subcommand::UpdateKeys => {
			commands::smith::update_session_keys(&data).await?;
		}
		Subcommand::SudoSetKey { new_key } => {
			data = data.build_keypair().build_client().await?;
			commands::sudo::set_key(data.keypair(), data.client(), new_key).await?;
		}
		Subcommand::ShowExpire { blocks, sessions } => {
			data = data.build_client().await?;
			commands::expire::monitor_expirations(&data, blocks, sessions).await?
		}
		Subcommand::ShowOnline => {
			data = data.build_client().await?;
			commands::smith::online(&data).await?
		}
	};

	Ok(())
}

/// rotate session keys
pub async fn rotate_keys(client: &Client) -> Result<SessionKeys, anyhow::Error> {
	client
		.rpc()
		.rotate_keys()
		.await?
		.deref()
		.try_into()
		.map_err(|e| anyhow!("Session keys have wrong length: {:?}", e))
}

/// set session keys
pub async fn set_session_keys(
	data: &Data,
	session_keys: SessionKeys,
) -> Result<TxProgress, subxt::Error> {
	data.client()
		.tx()
		.sign_and_submit_then_watch(
			&runtime::tx()
				.authority_members()
				.set_session_keys(session_keys),
			&PairSigner::new(data.keypair()),
			BaseExtrinsicParamsBuilder::new(),
		)
		.await
}

/// update session keys
pub async fn update_session_keys(data: &Data) -> Result<(), GcliError> {
	let session_keys = rotate_keys(data.client()).await?;
	let progress = set_session_keys(data, session_keys).await?;

	let _ = track_progress(progress).await?; // TODO
	Ok(())
}

/// submit go_online
pub async fn go_online(data: &Data) -> Result<(), GcliError> {
	if data
		.client()
		.storage()
		.fetch(
			&runtime::storage().session().next_keys(data.address()),
			None,
		)
		.await?
		.is_none()
	{
		return Err(GcliError::Logic(
			"This account has not set session keys!".to_string(),
		));
	}

	let progress = data.client()
		.tx()
		.sign_and_submit_then_watch(
			&runtime::tx().authority_members().go_online(),
			&PairSigner::new(data.keypair()),
			BaseExtrinsicParamsBuilder::new(),
		)
		.await?;

	let events = track_progress(progress).await?;
	if let Some(e) = events.find_first::<runtime::authority_members::events::MemberGoOnline>()? {
		println!("{e:?}");
	}
	Ok(())
}

/// submit go_offline
pub async fn go_offline(data: &Data) -> Result<(), subxt::Error> {
	let progress = data
		.client()
		.tx()
		.sign_and_submit_then_watch(
			&runtime::tx().authority_members().go_offline(),
			&PairSigner::new(data.keypair()),
			BaseExtrinsicParamsBuilder::new(),
		)
		.await?;
	let events = track_progress(progress).await?;
	if let Some(e) = events.find_first::<runtime::authority_members::events::MemberGoOffline>()? {
		println!("{e:?}");
	}

	Ok(())
}

/// get online authorities
pub async fn online(data: &Data) -> Result<(), anyhow::Error> {
	let client = data.client();
	let indexer = data.indexer.clone();

	let parent_hash = client
		.clone()
		.storage()
		.fetch(&runtime::storage().system().parent_hash(), None)
		.await?
		.unwrap();

	let mut identity_cache = cache::IdentityCache::new(client.clone(), indexer);

	let online_authorities = client
		.storage()
		.fetch(
			&runtime::storage().authority_members().online_authorities(),
			Some(parent_hash),
		)
		.await?
		.unwrap_or_default();

	println!("Online:");
	for identity_id in online_authorities {
		println!(
			"  {}",
			identity_cache
				.fetch_identity(identity_id, parent_hash)
				.await
				.unwrap_or_else(|_| format!("{identity_id}"))
		);
	}

	let incoming_authorities = client
		.storage()
		.fetch(
			&runtime::storage()
				.authority_members()
				.incoming_authorities(),
			Some(parent_hash),
		)
		.await?
		.unwrap_or_default();

	println!("Incoming:");
	for identity_id in incoming_authorities {
		println!(
			"  {}",
			identity_cache
				.fetch_identity(identity_id, parent_hash)
				.await
				.unwrap_or_else(|_| format!("{identity_id}"))
		);
	}

	let outgoing_authorities = client
		.storage()
		.fetch(
			&runtime::storage()
				.authority_members()
				.outgoing_authorities(),
			Some(parent_hash),
		)
		.await?
		.unwrap_or_default();

	println!("Outgoing:");
	for identity_id in outgoing_authorities {
		println!(
			"  {}",
			identity_cache
				.fetch_identity(identity_id, parent_hash)
				.await
				.unwrap_or_else(|_| format!("{identity_id}"))
		);
	}

	Ok(())
}

/// submit a certification and track progress
pub async fn cert(data: &Data, receiver: u32) -> Result<(), anyhow::Error> {
	let progress = data
		.client()
		.tx()
		.sign_and_submit_then_watch(
			&runtime::tx().smith_cert().add_cert(data.idty_index(), receiver),
			&PairSigner::new(data.keypair()),
			BaseExtrinsicParamsBuilder::new(),
		)
		.await?;

	let events = track_progress(progress).await?;

	// look for the expected event
	let new_cert_event = events.find_first::<runtime::smith_cert::events::NewCert>()?;
	let renew_cert_event = events.find_first::<runtime::smith_cert::events::RenewedCert>()?;

	if let Some(event) = new_cert_event {
		println!("{event:?}");
	}
	if let Some(event) = renew_cert_event {
		println!("{event:?}");
	}

	Ok(())
}
